#include "McKenzie.h"
#include "TargetDetection.h"
#include <math.h>
#include <Bluepad32.h>
#include "pitches.h"
#include <arduino-timer.h>

Timer<1, millis> IR_LED_Timer;

int tankNo = 1; //Assigned tank number (1 or 2) to prevent friendly fire

//Hardware Pin assignments
int targetFrontPin = A0;  //Analog pin for target 1 photo diode
int targetRearPin = A1;  //Analog pin for target 2 photo diode
int targetLeftPin = A2;  //Analog pin for target 3 photo diode
int targetRightPin = A3;  //Analog pin for target 4 photo diode

int pairPB = 0; //Pin that is connected to pair pushbutton
int buzzer = 10; //Pin connected to the piezo speaker
int fireLED = 2; //Pin connected to the infrared LED for firing
int pairLED = 3; //Pin connected to pair mode LED

int leftMotorDirA = 8; //Pin connected to the left motor controller direction A signal
int leftMotorDirB = 7; //Pin connected to the left motor controller direction B signal
int leftMotorPWM = 9; //Pin connected to the left motor controller PWM signal

int rightMotorDirA = 6; //Pin connected to the right motor controller direction A signal
int rightMotorDirB = 5; //Pin connected to the right motor controller direction B signal
int rightMotorPWM = 4; //Pin connected to the right motor controller PWM signal

int turretMotorDirA = 12; //Pin connected to the turret motor controller direction A signal
int turretMotorDirB = 11; //Pin connected to the turret motor controller direction B signal
int turretMotorPWM = 13; //Pin connected to the turret motor controller PWM signal

bool pairPBActive; //Pair pushbutton is being pressed
bool pairPBLastState; //State of Pair pushbutton for prior scan/loop  int controllerLeftStickPos = myControllers[gamepadNo]->axisY();

//Gamepad active variables
bool foundGamepad; //gamepad detected and active
int gamepadNo; //the number of the gamepad that is connected

//Drive Motor Variables
int controllerLeftStickPos; //Gamepad left stick position
int controllerRightStickPos; //Gamepad right stick position
float leftMotorPower; //Drive wheel left motor power
float rightMotorPower; //Drive wheel right wheel power
float drivePowerMax = 40.0; //Maximum Drive Motor Power (0-100)
  
//Turret Motor Variables
int controllerThrottlePos; //Gamepad throttle position
int controllerBrakePos; //Gamepad braker position
float turretMotorPower; //Turret motor power
float turretPowerMax = 30.0; //Maximum Turret Motor Power (0-100)

//Pair pushbutton variables
bool pairMode; //Pair mode active
Timer_rlm pairPBtimer (2000); //Pair mode pushbutton timer set to 2 seconds
Timer_rlm resetTimer (1000); //Game reset timer (2 seconds)
Timer_rlm onTime (500); //Timer for on-off flasher
Timer_rlm offTime (500); //Timer for on-off flasher
Timer_rlm firePulseTime (50); //Timer for pulsing fire LED on and off
Timer_rlm fireSoundDecay (20); //Timer for sound of tone decay when Fire() function is active
Timer_rlm hitSoundDelay (300); //Timer for delay between hit sound tones
Timer_rlm hitImmunity(5000);
Timer_rlm highPowerLimit(5000); //Limits use of high power mode to 5 seconds
Timer_rlm highPowerReset(5000); //Resets high power limit timer
Timer_rlm fireLimitReset(10000);

bool firePBactive; //one of the fire buttons on the controler is pressed
bool fireActive = false;

bool targetsActive; //one of the targets has been activated or hit 

int fireTone;

//variables for target averages
int targetLevelRight;
int targetLevelLeft;
int targetLevelFront;
int targetLevelRear;

int hitLevel = 5; //IR Target Level Change that indicates Hit 
int hitCycles = 3; //Number of cycles at frequency that have to be detected for hit
int hitInterval = 50; ////ON to ON and OFF to OFF interval of IR Pulse from other tank
int hitResetTime = 1000;

bool highPowerRequest = false;
int fireCount = 0; //Number of times fire sequence has been activated for controlling fire limits
int fireCountLimit = 3; //Number of fire sequences allowed within the fireLimitReset time

bool oneShot1 = false;
bool oneShot2 = false;
bool oneShot3 = false;
bool oneShot4 = false;
bool oneShot5 = false;
bool oneShot6 = false;
bool oneShot7 = false;
bool oneShot8 = false;
bool oneShot9 = false;

bool gameStarted = false;
bool gameEnded = false;
int hitCounter = 0; //Number of times the tank has been hit in a game
int gameResetCounter = 0;

Timer_rlm hitLightDelay(500); //Temporary timer for testing target IR detection

TargetDetection frontDetection(targetFrontPin, hitLevel, hitInterval, hitCycles, hitResetTime); //Instance of Target Detection Class for Front Target
TargetDetection rearDetection(targetRearPin, hitLevel, hitInterval, hitCycles, hitResetTime); //Instance of Target Detection Class for Rear Target
TargetDetection leftDetection(targetLeftPin, hitLevel, hitInterval, hitCycles, hitResetTime); //Instance of Target Detection Class for Left Target
TargetDetection rightDetection(targetRightPin, hitLevel, hitInterval, hitCycles, hitResetTime); //Instance of Target Detection Class for Right Target

ControllerPtr myControllers[BP32_MAX_CONTROLLERS];
bool IR_Timer_Handler(void *)
{
  if (fireActive) {
  //digitalWrite(fireLED, !digitalRead(fireLED)); // Toggle IR LED
  }
  //else {digitalWrite(fireLED, LOW);}
  return true; // repeat? true
}

int galagaNotes[] = {
  NOTE_E5, NOTE_C5, NOTE_D5, NOTE_F5, NOTE_E5, NOTE_C5, NOTE_D5, NOTE_A5, 
  NOTE_G5, NOTE_C5, NOTE_D5, NOTE_F4, NOTE_E5, NOTE_C5, NOTE_G5, NOTE_B5,
  NOTE_C6, NOTE_AS5, NOTE_A5, NOTE_G5, NOTE_F5, NOTE_E5, NOTE_D5, NOTE_D5,
  NOTE_AS5, NOTE_C6, NOTE_B5, NOTE_G5, NOTE_A5, NOTE_F5, NOTE_D5, NOTE_G5, NOTE_E5, NOTE_D5
};

int galagaNotes_OLD[] = {
  NOTE_E3, NOTE_C3, NOTE_D3, NOTE_F3, NOTE_E3, NOTE_C3, NOTE_D3, NOTE_A3, 
  NOTE_G3, NOTE_C3, NOTE_D3, NOTE_F2, NOTE_E3, NOTE_C3, NOTE_G3, NOTE_B3,
  NOTE_C4, NOTE_AS3, NOTE_A3, NOTE_G3, NOTE_F3, NOTE_E3, NOTE_E3, NOTE_D3,
  NOTE_AS3, NOTE_C4, NOTE_B3, NOTE_G3, NOTE_A3, NOTE_F3, NOTE_D3, NOTE_G3, NOTE_E3, NOTE_D3
};

int galagaDurations[] = {
  6, 16, 6, 16, 6, 16, 6, 16, 
  6, 16, 6, 16, 6, 16, 6, 16,
  6, 16, 6, 16, 6, 16, 6, 16,
  6, 16, 6, 16, 12, 12, 12, 12, 12, 12
};

int tapsNotes[] = {
  NOTE_G3, NOTE_G3, 
  NOTE_C4, NOTE_G3, NOTE_C4, 
  NOTE_E4, NOTE_G3, NOTE_C4, 
  NOTE_E4, NOTE_G3, NOTE_C4, NOTE_E4, NOTE_G3, NOTE_C4, 
  NOTE_E4, NOTE_C4, NOTE_E4, 
  NOTE_G4, NOTE_E4, NOTE_C4,
  NOTE_G3, NOTE_G3, NOTE_G3,
  NOTE_C4
};

float tapsDurations[] = {
  6, 16, 
  1.5, 6, 16, 
  1.5, 8, 8, 
  4, 8, 8, 4, 8, 8,
  1.5, 8, 8,
  2, 4, 4, 
  1.5, 6, 16,
  1.5
};

int marioBrosNotes[]= {
  NOTE_E5, NOTE_E5, REST, NOTE_E5, REST, NOTE_C5, NOTE_E5,
  NOTE_G5, REST, NOTE_G4, REST, 
  NOTE_C5, NOTE_G4, REST, NOTE_E4,
  NOTE_A4, NOTE_B4, NOTE_AS4, NOTE_A4,
  NOTE_G4, NOTE_E5, NOTE_G5, NOTE_A5, NOTE_F5, NOTE_G5,
  REST, NOTE_E5,NOTE_C5, NOTE_D5, NOTE_B4,
  NOTE_C5, NOTE_G4, REST, NOTE_E4,
  NOTE_A4, NOTE_B4, NOTE_AS4, NOTE_A4,
  NOTE_G4, NOTE_E5, NOTE_G5, NOTE_A5, NOTE_F5, NOTE_G5,
  REST, NOTE_E5,NOTE_C5, NOTE_D5, NOTE_B5,
  
  REST, NOTE_G5, NOTE_FS5, NOTE_F5, NOTE_DS5, NOTE_E5,
  REST, NOTE_GS4, NOTE_A4, NOTE_C4, REST, NOTE_A4, NOTE_C5, NOTE_D5,
  REST, NOTE_DS5, REST, NOTE_D5,
  NOTE_C5, REST,
  
  REST, NOTE_G5, NOTE_FS5, NOTE_F5, NOTE_DS5, NOTE_E5,
  REST, NOTE_GS4, NOTE_A4, NOTE_C4, REST, NOTE_A4, NOTE_C5, NOTE_D5,
  REST, NOTE_DS5, REST, NOTE_D5,
  NOTE_C5, REST,
  
  NOTE_C5, NOTE_C5, NOTE_C5, REST, NOTE_C5, NOTE_D5,
  NOTE_E5, NOTE_C5, NOTE_A4, NOTE_G4,
  
  NOTE_C5, NOTE_C5, NOTE_C5, REST, NOTE_C5, NOTE_D5, NOTE_E5,
  REST, 
  NOTE_C5, NOTE_C5, NOTE_C5, REST, NOTE_C5, NOTE_D5,
  NOTE_E5, NOTE_C5, NOTE_A4, NOTE_G4,
  NOTE_E5, NOTE_E5, REST, NOTE_E5, REST, NOTE_C5, NOTE_E5,
  NOTE_G5, REST, NOTE_G4, REST, 
  NOTE_C5, NOTE_G4, REST, NOTE_E4,
  
  NOTE_A4, NOTE_B4, NOTE_AS4, NOTE_A4,
  NOTE_G4, NOTE_E5, NOTE_G5, NOTE_A5, NOTE_F5, NOTE_G5,
  REST, NOTE_E5, NOTE_C5, NOTE_D5, NOTE_B4,
  
  NOTE_C5, NOTE_G4, REST, NOTE_E4,
  NOTE_A4, NOTE_B4, NOTE_AS4, NOTE_A4,
  NOTE_G4, NOTE_E5, NOTE_G5, NOTE_A5, NOTE_F5, NOTE_G5,
  REST, NOTE_E5, NOTE_C5, NOTE_D5, NOTE_B4,
  
  NOTE_E5, NOTE_C5, NOTE_G4, REST, NOTE_GS4,
  NOTE_A4, NOTE_F5, NOTE_F5, NOTE_A4,
  NOTE_D5, NOTE_A5, NOTE_A5, NOTE_A5, NOTE_G5, NOTE_F5,
  
  NOTE_E5, NOTE_C5, NOTE_A4, NOTE_G4,
  NOTE_E5, NOTE_C5, NOTE_G4, REST, NOTE_GS4,
  NOTE_A4, NOTE_F5, NOTE_F5, NOTE_A4,
  NOTE_B4, NOTE_F5, NOTE_F5, NOTE_F5, NOTE_E5, NOTE_D5,
  NOTE_C5, NOTE_E4, NOTE_E4, NOTE_C4,
  
  NOTE_E5, NOTE_C5, NOTE_G4, REST, NOTE_GS4,
  NOTE_A4, NOTE_F5, NOTE_F5, NOTE_A4,
  NOTE_D5, NOTE_A5, NOTE_A5, NOTE_A5, NOTE_G5, NOTE_F5,
  
  NOTE_E5, NOTE_C5, NOTE_A4, NOTE_G4,
  NOTE_E5, NOTE_C5, NOTE_G4, REST, NOTE_GS4,
  NOTE_A4, NOTE_F5, NOTE_F5, NOTE_A4,
  NOTE_B4, NOTE_F5, NOTE_F5, NOTE_F5, NOTE_E5, NOTE_D5,
  NOTE_C5, NOTE_E4, NOTE_E4, NOTE_C4,
  NOTE_C5, NOTE_C5, NOTE_C5, REST, NOTE_C5, NOTE_D5, NOTE_E5,
  REST,
  
  NOTE_C5, NOTE_C5, NOTE_C5, REST, NOTE_C5, NOTE_D5,
  NOTE_E5, NOTE_C5, NOTE_A4, NOTE_G4,
  NOTE_E5, NOTE_E5, REST, NOTE_E5, REST, NOTE_C5, NOTE_E5,
  NOTE_G5, REST, NOTE_G4, REST, 
  NOTE_E5, NOTE_C5, NOTE_G4, REST, NOTE_GS4,
  NOTE_A4, NOTE_F5, NOTE_F5, NOTE_A4,
  NOTE_D5, NOTE_A5, NOTE_A5, NOTE_A5, NOTE_G5, NOTE_F5,
  
  NOTE_E5, NOTE_C5, NOTE_A4, NOTE_G4,
  NOTE_E5, NOTE_C5, NOTE_G4, REST, NOTE_GS4,
  NOTE_A4, NOTE_F5, NOTE_F5, NOTE_A4,
  NOTE_B4, NOTE_F5, NOTE_F5, NOTE_F5, NOTE_E5, NOTE_D5,
  NOTE_C5, NOTE_E4, NOTE_E4, NOTE_C4,

};

int marioBrosDurations[]={
  8, 8, 8, 8, 8, 8, 8,
  4, 4, 8, 4, 
  4, 8, 4, 4,
  4, 4, 8, 4,
  8, 8, 8, 4, 8, 8,
  8, 4,8, 8, 4,
  4, 8, 4, 4,
  4, 4, 8, 4,
  8, 8, 8, 4, 8, 8,
  8, 4,8, 8, 4,
  
  
  4, 8, 8, 8, 4, 8,
  8, 8, 8, 8, 8, 8, 8, 8,
  4, 4, 8, 4,
  2, 2,
  
  4, 8, 8, 8, 4, 8,
  8, 8, 8, 8, 8, 8, 8, 8,
  4, 4, 8, 4,
  2, 2,
  
  8, 4, 8, 8, 8, 4,
  8, 4, 8, 2,
  
  8, 4, 8, 8, 8, 8, 8,
  1, 
  8, 4, 8, 8, 8, 4,
  8, 4, 8, 2,
  8, 8, 8, 8, 8, 8, 4,
  4, 4, 4, 4, 
  4, 8, 4, 4,
  
  4, 4, 8, 4,
  8, 8, 8, 4, 8, 8,
  8, 4, 8, 8, 4,
  
  4, 8, 4, 4,
  4, 4, 8, 4,
  8, 8, 8, 4, 8, 8,
  8, 4, 8, 8, 4,
  
  8, 4, 8, 4, 4,
  8, 4, 8, 2,
  8, 8, 8, 8, 8, 8,
  
  8, 4, 8, 2,
  8, 4, 8, 4, 4,
  8, 4, 8, 2,
  8, 4, 8, 8, 8, 8,
  8, 4, 8, 2,
  
  8, 4, 8, 4, 4,
  8, 4, 8, 2,
  8, 8, 8, 8, 8, 8,
  
  8, 4, 8, 2,
  8, 4, 8, 4, 4,
  8, 4, 8, 2,
  8, 4, 8, 8, 8, 8,
  8, 4, 8, 2,
  8, 4, 8, 8, 8, 8, 8,
  1,
  
  8, 4, 8, 8, 8, 4,
  8, 4, 8, 2,
  8, 8, 8, 8, 8, 8, 4,
  4, 4, 4, 4, 
  8, 4, 8, 4, 4,
  8, 4, 8, 2,
  8, 8, 8, 8, 8, 8,
  
  8, 4, 8, 2,
  8, 4, 8, 4, 4,
  8, 4, 8, 2,
  8, 4, 8, 8, 8, 8,
  8, 4, 8, 2,
  
};

// Arduino setup function. Runs in CPU 1
void setup() {
  if (tankNo == 1){
    firePulseTime.preset = 30; 
    frontDetection.hitInterval = 76;
    rearDetection.hitInterval = 76;
    leftDetection.hitInterval = 76;
    rightDetection.hitInterval = 76;
    IR_LED_Timer.every(15, IR_Timer_Handler);

  } 
  else {
    firePulseTime.preset = 76;
    frontDetection.hitInterval = 30;
    rearDetection.hitInterval = 30;
    leftDetection.hitInterval = 30;
    rightDetection.hitInterval = 30;
    IR_LED_Timer.every(38, IR_Timer_Handler);
  }
  
  //Digital I/O Setup
  pinMode (pairPB, INPUT);
  pinMode (buzzer, OUTPUT);
  pinMode (pairLED, OUTPUT);
  pinMode (fireLED, OUTPUT);

  pinMode (leftMotorDirA, OUTPUT);
  pinMode (leftMotorDirB, OUTPUT);
  pinMode (leftMotorPWM, OUTPUT);

  pinMode (rightMotorDirA, OUTPUT);
  pinMode (rightMotorDirB, OUTPUT);
  pinMode (rightMotorPWM, OUTPUT);

  pinMode (turretMotorDirA, OUTPUT);
  pinMode (turretMotorDirB, OUTPUT);
  pinMode (turretMotorPWM, OUTPUT);

  // Initialize serial
  //Serial.begin(1200);
  /* while (!Serial) {
    // wait for serial port to connect.
    // You don't have to do this in your game. This is only for debugging
    // purposes, so that you can see the output in the serial console.
    ;
  }
  */

  String fv = BP32.firmwareVersion();
  Serial.print("Firmware version installed: ");
  Serial.println(fv);

  // To get the BD Address (MAC address) call:
  const uint8_t* addr = BP32.localBdAddress();
  Serial.print("BD Address: ");
  for (int i = 0; i < 6; i++) {
    //Serial.print(addr[i], HEX);
    if (i < 5)
      Serial.print(":");
    else
      Serial.println();
  }

  // BP32.pinMode(27, OUTPUT);
  // BP32.digitalWrite(27, 0);

  // This call is mandatory. It setups Bluepad32 and creates the callbacks.
  BP32.setup(&onConnectedController, &onDisconnectedController);

  // "forgetBluetoothKeys()" should be called when the user performs
  // a "device factory reset", or similar.
  // Calling "forgetBluetoothKeys" in setup() just as an example.
  // Forgetting Bluetooth keys prevents "paired" gamepads to reconnect.
  // But might also fix some connection / re-connection issues.
  // BP32.forgetBluetoothKeys();
  
  pairPBtimer.start();
  onTime.start();
  delay(1000);
  onTime.start();
  firePulseTime.reset();
  firePulseTime.start();
}

// This callback gets called any time a new gamepad is connected.
// Up to 4 gamepads can be connected at the same time.
void onConnectedController(ControllerPtr ctl) {
  bool foundEmptySlot = false;
  for (int i = 0; i < BP32_MAX_GAMEPADS; i++) {
    if (myControllers[i] == nullptr) {
      Serial.print("CALLBACK: Controller is connected, index=");
      Serial.println(i);
      myControllers[i] = ctl;
      foundEmptySlot = true;

      // Optional, once the gamepad is connected, request further info about the
      // gamepad.
      ControllerProperties properties = ctl->getProperties();
      char buf[80];
      sprintf(buf,
              "BTAddr: %02x:%02x:%02x:%02x:%02x:%02x, VID/PID: %04x:%04x, "
              "flags: 0x%02x",
              properties.btaddr[0], properties.btaddr[1], properties.btaddr[2],
              properties.btaddr[3], properties.btaddr[4], properties.btaddr[5],
              properties.vendor_id, properties.product_id, properties.flags);
      Serial.println(buf);
      break;
    }
  }
  if (!foundEmptySlot) {
    Serial.println(
        "CALLBACK: Controller connected, but could not found empty slot");
  }
}

void onDisconnectedController(ControllerPtr ctl) {
  bool foundGamepad = false;

  for (int i = 0; i < BP32_MAX_GAMEPADS; i++) {
    if (myControllers[i] == ctl) {
      Serial.print("CALLBACK: Controller is disconnected from index=");
      Serial.println(i);
      myControllers[i] = nullptr;
      foundGamepad = true;
      break;
    }
  }

  if (!foundGamepad) {
    Serial.println(
        "CALLBACK: Controller disconnected, but not found in myControllers");
  }
}

void processGamepad(ControllerPtr gamepad) {
  // There are different ways to query whether a button is pressed.
  // By query each button individually:
  //  a(), b(), x(), y(), l1(), etc...

  if (gamepad->a()) {
    static int colorIdx = 0;
    // Some gamepads like DS4 and DualSense support changing the color LED.
    // It is possible to change it by calling:
    switch (colorIdx % 3) {
      case 0:
        // Red
        gamepad->setColorLED(255, 0, 0);
        break;
      case 1:
        // Green
        gamepad->setColorLED(0, 255, 0);
        break;
      case 2:
        // Blue
        gamepad->setColorLED(0, 0, 255);
        break;
    }
    colorIdx++;
  }

  if (gamepad->b()) {
    // Turn on the 4 LED. Each bit represents one LED.
    static int led = 0;
    led++;
    // Some gamepads like the DS3, DualSense, Nintendo Wii, Nintendo Switch
    // support changing the "Player LEDs": those 4 LEDs that usually indicate
    // the "gamepad seat".
    // It is possible to change them by calling:
    gamepad->setPlayerLEDs(led & 0x0f);
  }

  if (gamepad->x()) {
    // Duration: 255 is ~2 seconds
    // force: intensity
    // Some gamepads like DS3, DS4, DualSense, Switch, Xbox One S support
    // rumble.
    // It is possible to set it by calling:
    gamepad->setRumble(0xc0 /* force */, 0xc0 /* duration */);
  }

  // Another way to query the buttons, is by calling buttons(), or
  // miscButtons() which return a bitmask.
  // Some gamepads also have DPAD, axis and more.
  char buf[256];
  snprintf(buf, sizeof(buf) - 1,
           "idx=%d, dpad: 0x%02x, buttons: 0x%04x, "
           "axis L: %4li, %4li, axis R: %4li, %4li, "
           "brake: %4ld, throttle: %4li, misc: 0x%02x, "
           "gyro x:%6d y:%6d z:%6d, accel x:%6d y:%6d z:%6d, "
           "battery: %d",
           gamepad->index(),        // Gamepad Index
           gamepad->dpad(),         // DPAD
           gamepad->buttons(),      // bitmask of pressed buttons
           gamepad->axisX(),        // (-511 - 512) left X Axis
           gamepad->axisY(),        // (-511 - 512) left Y axis
           gamepad->axisRX(),       // (-511 - 512) right X axis
           gamepad->axisRY(),       // (-511 - 512) right Y axis
           gamepad->brake(),        // (0 - 1023): brake button
           gamepad->throttle(),     // (0 - 1023): throttle (AKA gas) button
           gamepad->miscButtons(),  // bitmak of pressed "misc" buttons
           gamepad->gyroX(),      // Gyro X
           gamepad->gyroY(),      // Gyro Y
           gamepad->gyroZ(),      // Gyro Z
           gamepad->accelX(),     // Accelerometer X
           gamepad->accelY(),     // Accelerometer Y
           gamepad->accelZ(),     // Accelerometer Z
           gamepad->battery()       // 0=Unknown, 1=empty, 255=full
  );
  //Serial.println(buf);

  // You can query the axis and other properties as well. See
  // Controller.h For all the available functions.
}

void MotorOutput (float motorPower, int dirAPin, int dirBPin, int pwmPin) {
  // MotorOutput Function 
  // Controls digital and analog outputs connected to
  // Motor Controller such as the L298N
  // 100 = Max power in forward direction
  // -100 = Max power in reverse direction
  
  //Paremeters
  //  motorPower - requeseted power
  //  dirAPin - digital pin connected to the motor controller direction A input
  //  dirBPin - digital pin connected to the motor controller direction A input
  //  pwmPin  - digital pin connected to the motor controller pwm input

  if (motorPower > 1) {
    digitalWrite(dirAPin, HIGH);
    digitalWrite(dirBPin, LOW);
  }
  else if (motorPower < 1){   
    digitalWrite(dirAPin, LOW);
    digitalWrite(dirBPin, HIGH);
  }
  else {
    digitalWrite(dirAPin, LOW);
    digitalWrite(dirBPin, LOW);
  } 
  long power = abs (motorPower)*10;
  int pwmOutput = map(power, 0,1000, 0, 255);  //Scales motor power to 0-255 for PWM output
  analogWrite (pwmPin, pwmOutput);
}

void Fire(){
  firePulseTime.reset();
  firePulseTime.start();
  fireSoundDecay.reset();
  fireSoundDecay.start();
  fireTone = 2500;
  fireActive = true;
  
  while (fireTone>1500){
    
    firePulseTime.check();
    int PulseOnTime = firePulseTime.preset / 2;
    if (firePulseTime.acc < PulseOnTime) {
      digitalWrite(fireLED,HIGH);
    }
    else {
      digitalWrite(fireLED,LOW);
    }
    
    if (firePulseTime.dn){
      firePulseTime.reset();
      firePulseTime.start();
    }
    
    //IR_LED_Timer.tick(); // tick the timer
    
    fireSoundDecay.check();
    if (fireSoundDecay.dn){
      tone(buzzer, fireTone);
      fireTone = fireTone - 7;
      fireSoundDecay.reset();
      fireSoundDecay.start();
    }
    GamePadCheck();
    MotorControl();
    
    
    if (frontDetection.check() || rearDetection.check() || leftDetection.check() || rightDetection.check()){
      hitLightDelay.reset();
      hitLightDelay.start();
      delay(150);
      noTone(buzzer);
      fireActive = false;
      return;  
    }
    
  }
  delay(150);
  noTone(buzzer);     // Stop sound...
  digitalWrite(fireLED,LOW);
  fireActive = false;
}

void Hit(){
  int i = 1;
  hitSoundDelay.reset();
  hitSoundDelay.start();
  int toneDuration = hitSoundDelay.preset - 50;
  tone(buzzer,2200,toneDuration);
  bool highLow = false;
  
  while (i<=8){
    hitSoundDelay.check();
    if (hitSoundDelay.dn){
      hitSoundDelay.reset();
      hitSoundDelay.start();
      i++;
      if (highLow) {
        highLow=false;
        tone(buzzer, 2200, toneDuration);
      }
      else {
        highLow=true;
        tone(buzzer, 1800, toneDuration);
      }
    }
    GamePadCheck();
    MotorControl();
  }
  pacManDeath();
  noTone(buzzer);
  delay(200);
}

void Hit_OLD2(){
  for (int i = 1; i<5; i++){
    tone(buzzer, 2200,500);
    delay (550);
    tone(buzzer, 1800,500);
    delay(550);
  }
  noTone(buzzer);
  delay(200);
}

void Hit_OLD(){
  for (int j = 1500 ; j>1400; j=j-2){
    for (int i = 1 ; i<7 ; i++) {
      tone(buzzer, j,4);
      delay(5);
      tone(buzzer, j-30,4);
      delay(5);
    }  
  }
  noTone(buzzer); 
  delay(10);
  for (int i = 1 ; i<50 ; i++) {
      tone(buzzer, 1400,4);
      delay(5);
      tone(buzzer, 1350,4);
      delay(5);
  }
  delay(10);
  noTone(buzzer);  
  
  delay(150);
}

void GalagaMusic(){
  int size = sizeof (galagaDurations)/4;
  for (int thisNote = 0; thisNote < size; thisNote++) {
    // to calculate the note duration, take one second divided by the note type.
    //e.g. quarter note = 1000 / 4, eighth note = 1000/8, etc.
    int noteDuration = 1500 / galagaDurations[thisNote];
    //int noteDuration = 2500 / galagaDurations[thisNote];
    tone(buzzer, galagaNotes[thisNote], noteDuration);

    // to distinguish the notes, set a minimum time between them.
    // the note's duration + 30% seems to work well:
    int pauseBetweenNotes = 10 + noteDuration * 1.10;
    delay(pauseBetweenNotes);
    // stop the tone playing:
  }
  delay(1000);
  noTone(buzzer);
  delay(10);
}

void MarioBrosMusic(){
  int size = sizeof (marioBrosDurations)/4;
  for (int thisNote = 0; thisNote < size; thisNote++) {
    // to calculate the note duration, take one second divided by the note type.
    //e.g. quarter note = 1000 / 4, eighth note = 1000/8, etc.
    int noteDuration = 1000 / marioBrosDurations[thisNote];
    tone(buzzer, marioBrosNotes[thisNote], noteDuration);

    // to distinguish the notes, set a minimum time between them.
    // the note's duration + 30% seems to work well:
    int pauseBetweenNotes = 10 + noteDuration * 1.15;
    delay(pauseBetweenNotes);

    //delay(150);
    // stop the tone playing:
    noTone(buzzer);
    delay (10);

  }
}

void tapsMusic(){
  int size = sizeof (tapsDurations)/4;
  for (int thisNote = 0; thisNote < size; thisNote++) {
    // to calculate the note duration, take one second divided by the note type.
    //e.g. quarter note = 1000 / 4, eighth note = 1000/8, etc.
    int noteDuration = 2000 / tapsDurations[thisNote];
    tone(buzzer, tapsNotes[thisNote], noteDuration);

    // to distinguish the notes, set a minimum time between them.
    // the note's duration + 30% seems to work well:
    int pauseBetweenNotes = 10 + noteDuration * 1.15;
    delay(pauseBetweenNotes);

    //delay(150);
    // stop the tone playing:
    noTone(buzzer);
    delay (10);

  }
}

void GamePadCheck(){
  // This call fetches all the controller info from the NINA (ESP32) module.
  // The controllers pointer (the ones received in the callbacks) gets updated
  // automatically.

  BP32.update();

  // It is safe to always do this before using the controller API.
  // This guarantees that the controller is valid and connected.
  for (int i = 0; i < BP32_MAX_CONTROLLERS; i++) {
    ControllerPtr myController = myControllers[i];

    if (myController && myController->isConnected()) {
      //processGamepad(myController);
      foundGamepad = true;
      gamepadNo = i;
      break;
    }
    else {
      foundGamepad = false;
    }
  }
}

void MotorControl() {
   int controllerButtons = myControllers[gamepadNo]->buttons();

  if (foundGamepad && !gameEnded) {
    MotorOutput (leftMotorPower, leftMotorDirA, leftMotorDirB, leftMotorPWM);
    MotorOutput (rightMotorPower, rightMotorDirA, rightMotorDirB, rightMotorPWM);
    MotorOutput (turretMotorPower, turretMotorDirA, turretMotorDirB, turretMotorPWM);
  }
  else {
    MotorOutput (0, leftMotorDirA, leftMotorDirB, leftMotorPWM);
    MotorOutput (0, rightMotorDirA, rightMotorDirB, rightMotorPWM);
    MotorOutput (0, turretMotorDirA, turretMotorDirB, turretMotorPWM);
  }

  //Left and Right Movement
  controllerLeftStickPos = myControllers[gamepadNo]->axisY();
  controllerRightStickPos = myControllers[gamepadNo]->axisRY();
  leftMotorPower = (controllerLeftStickPos / 512.0) * drivePowerMax;
  rightMotorPower = -(controllerRightStickPos / 512.0) * drivePowerMax;

  //Turret Movement
  controllerThrottlePos = myControllers[gamepadNo]->throttle();
  controllerBrakePos = myControllers[gamepadNo]->brake();
  turretMotorPower = ((controllerThrottlePos - controllerBrakePos) / 1023.0) * turretPowerMax;

  //High Drive Speed
  if (controllerButtons & 0b100000) {highPowerRequest = true;}
  else {highPowerRequest=false;}
  
  //High Power limit - can only use high power for 5 seconds in a 20 second window
  highPowerLimit.check();
  if (highPowerRequest && !highPowerLimit.dn){
  //if (highPowerRequest){
    if (!oneShot6) {
      highPowerLimit.reset();
      oneShot6 = true;
    }
    highPowerLimit.start();
    drivePowerMax = 100.0;
  }
  else {
    drivePowerMax = 40.0;
    highPowerLimit.stop();
  }
  
  if (highPowerLimit.dn && !oneShot7) {
    oneShot7 = true;
    highPowerReset.reset();
    highPowerReset.start();
  }
  highPowerReset.check();
  if (highPowerReset.dn){
    highPowerLimit.reset();
    oneShot6 = false;
    oneShot7 = false;
    highPowerReset.reset();
  }
  
  //Slow Turret Speed
  if (controllerButtons & 0b010000) {
    turretPowerMax = 15.0;
  }
  else {
    turretPowerMax = 30.0;
  }
}

void PairMode(){
  //Pair Pushbutton action
  pairPBActive = digitalRead(pairPB);
  
  if (pairPBActive) {
    pairPBtimer.check();
  }
  else {
    pairPBtimer.reset();
    pairPBtimer.start();
  }
  pairPBtimer.check(); 
  if (pairPBtimer.dn) {
    if (!pairMode) {
      pairMode=true;
    }
    else {pairMode=false;}
    pairPBtimer.reset();
    pairPBtimer.start();
  }
  if (pairMode) {
    BP32.forgetBluetoothKeys();
    BP32.enableNewBluetoothConnections(false);
    }
  else {BP32.enableNewBluetoothConnections(true);}
}

void GameReset(){
  //Pair Pushbutton for Game Reset looks for 3 pushbutton presses within 2 seconds
  pairPBActive = digitalRead(pairPB);
  
  resetTimer.check();
  if (pairPBActive && !oneShot2) {
    oneShot2 = true;
    gameResetCounter++;
    if (!resetTimer.tt){
      resetTimer.reset();
      resetTimer.start();
    }
  }
  if (!pairPBActive) {
    oneShot2 = false;
  }
  if (resetTimer.dn){
    gameResetCounter = 0;
  }
  if (gameResetCounter >= 3){
    gameStarted = false;
    gameEnded = false;
    hitCounter = 0;
  }
}

void pacManDeath(){
  int start;
  int end;
  for (int i=1; i<=4; i++){
    start = 2500 - i*350;
    end = start + 800;
    for (int j=start; j<end; j=j+70){
      tone(buzzer,j,10);
      delay(12);
      GamePadCheck();
      MotorControl();
    }
    for (int j=end; j>start; j=j-70){
      tone(buzzer,j,10);
      delay(12);
      GamePadCheck();
      MotorControl();
    }
  }
  delay(150);
  for (int i=1; i<=2; i++){
    for (int j=2000; j<2800; j=j+70){
      tone(buzzer,j,10);
      delay(12);
      GamePadCheck();
      MotorControl();
    }
  }
  delay (100);
  noTone (buzzer);
}


// Arduino loop function. Runs in CPU 1
void loop() {
  //IR_LED_Timer.tick(); // tick the timer
  
  //Serial.println(firePulseTime.preset);
  if (!gameStarted){
    gameStarted = true;
    gameResetCounter = 0;
    GalagaMusic();
  }

  GamePadCheck();
  if (!gameEnded) {MotorControl();}
  //PairMode();
  GameReset();
  
  //Serial.print(analogRead(A0));
  //Serial.print("  ");
  //Serial.print(analogRead(A1));
  //Serial.print("  ");
  //Serial.print(analogRead(A2));
  //Serial.print("  ");
  //Serial.println(analogRead(A3));
  
  //if (frontDetection.check() || rearDetection.check() || leftDetection.check() || rightDetection.check()){ 
  bool detected = (frontDetection.check() || rearDetection.check() || leftDetection.check() || rightDetection.check());    
  if (detected){
    hitLightDelay.reset();
    hitLightDelay.start();
    //hitImmunity.reset();
    //hitImmunity.start();
  }

  hitLightDelay.check();
  if (hitLightDelay.tt) {targetsActive=true;}
  else {targetsActive=false;}

  hitImmunity.check();
  if (targetsActive && !oneShot1 && !hitImmunity.tt && !gameEnded){
    myControllers[gamepadNo]->setRumble(0xc0, 0xc0);
    // delay(500);
    Hit();
    hitImmunity.reset();
    hitImmunity.start();
    hitCounter++;
    oneShot1 = true;
    //digitalWrite(pairLED,HIGH);
  }
  if (!targetsActive){
    oneShot1 = false;
    //digitalWrite(pairLED,LOW);
  }

  if ((hitCounter > 3) && !gameEnded){
    gameEnded = true;
    //pacManDeath();
    tapsMusic();
  }

  //Fire Pushbutton
  int controllerButtons = myControllers[gamepadNo]->buttons();
  if ((controllerButtons & 0b100000)&&(controllerButtons & 0b10000)&&!gameEnded) {firePBactive=true;}
  else {firePBactive=false;}
  if (firePBactive && !gameEnded && !oneShot5 && (fireCount < fireCountLimit)){
    oneShot5 = true;
    Fire();
    fireCount++;
  }
  else {
    //digitalWrite(fireLED,LOW); 
  }
  if (!firePBactive){
    oneShot5 = false;
  }
  if ((fireCount>0) && !oneShot8){
    fireLimitReset.reset();
    fireLimitReset.start();
    oneShot8 = true;
  }
  fireLimitReset.check();
  if (fireLimitReset.dn){
    fireCount = 0;
    oneShot8 = false;
  }

  // on/off pulse timer
  onTime.check();
  if (onTime.dn) {
    offTime.start();
    onTime.reset();
  }
  offTime.check();
  if (offTime.dn){
    offTime.reset();
    onTime.start();
  }

  if (onTime.tt || foundGamepad){
    digitalWrite(pairLED,HIGH);
  }
  else {
    digitalWrite(pairLED, LOW);  
  }
}